﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using RobotBehaviourEditor.Model;
using System.IO;
using System.Windows.Forms;

namespace RobotBehaviourEditor.Formats
{

    /**
     * A class rensponsible for parsing a List of JointMotions out
     * of an XML in a Dream Team (FIIT team) format
     * 
     * */
    public class XmlImporter : RobotBehaviourEditor.Formats.IXmlImporter
    {

        private string filePath;

        private List<PhaseNode> allPhases = new List<PhaseNode>();

        private string initialPhase = null;

        private string lowLevelSkill = null;

        public XmlImporter(string filePath)
        {
            this.filePath = filePath;
        }

        public List<string> lowLevelSkillsInFile()
        {
            XmlReader reader = createXmlReader();
            List<string> lowSkillNames = lowSkills(reader);

            if (lowSkillNames.Count == 0)
                throw new Exception("V súbore sa nenašiel žiaden low_skill tag");

            reader.Close();
            return lowSkillNames;
        }

        private XmlReader createXmlReader()
        {
            FileInfo xmlFile = new FileInfo(filePath);
            XmlReader reader = new XmlTextReader("file://" + xmlFile.FullName);
            return reader;
        }

        private List<string> lowSkills(XmlReader reader)
        {
            List<string> lowSkillNames = new List<string>();

            while (reader.Read())
                if (reader.NodeType == XmlNodeType.Element && reader.Name.Equals("low_skill"))
                    while (reader.MoveToNextAttribute())
                        if ("name".Equals(reader.Name))
                            lowSkillNames.Add(reader.Value);
            return lowSkillNames;
        }

        public List<JointMotion> movesOf(string lowLevelSkill)
        {
            XmlReader reader = createXmlReader();
            this.lowLevelSkill = lowLevelSkill;
            PhaseNode phase = null;

            while (reader.Read())
            {
                if (reader.NodeType == XmlNodeType.EndElement)
                    continue;
                if ("low_skill".Equals(reader.Name))
                    findInitialPhase(reader);

                if(reader.Name.Equals("phase"))
                {
                    phase = PhaseNode.parse(reader);
                    allPhases.Add(phase);
                 }
                if (reader.Name.Equals("speed_constant"))
                    phase.speedConstant = parseSpeedConstant(reader);

                if (reader.Name.Equals("efector"))
                    phase.addMove(EffectorNode.parse(reader) );
           }
            List<JointMotion> motions = new List<JointMotion>();
            double elapsedTime = 0.0;

            foreach (PhaseNode aPhase in phasesOfChosenSkill())
            {
                // time = toRadian(angleDifference) / (angleDifference*speed_constant / 1000) => time = 17.45 / speed_c
                double durationOfThisPhase = (17450 / aPhase.speedConstant);
                foreach (EffectorNode move in aPhase.getMoves())
                {
                    JointMotion motion = new JointMotion();
                    motion.Joint = Joints.GetJointTypeById(move.jointName);
                    motion.StartPosition = move.startAngle;
                    motion.EndPosition = move.endAngle;
                    motion.StartTime = (int)elapsedTime;
                    motion.Duration = (int)durationOfThisPhase;
                    motions.Add(motion);
                 }
                elapsedTime = elapsedTime + durationOfThisPhase;
            }

            return motions;
       }              

        private void findInitialPhase(XmlReader reader)
        {
            reader.MoveToNextAttribute();
            if (lowLevelSkill.Equals(reader.Value))
            {
                while (reader.NodeType != XmlNodeType.Element)
                    reader.Read();
                reader.MoveToNextAttribute();
                initialPhase = reader.Value;
            }
        }

        private int parseSpeedConstant(XmlReader reader)
        {
            reader.Read();
            return int.Parse(reader.Value);
        }  

        private List<PhaseNode> phasesOfChosenSkill()
        {
            List<PhaseNode> phasesOfTheSkill = new List<PhaseNode>();
            PhaseNode currentPhase = findNodeWithNameIn(initialPhase, allPhases);

            while (!currentPhase.isFinal)
            {
                phasesOfTheSkill.Add(currentPhase);
                currentPhase = findNodeWithNameIn(currentPhase.nextName, allPhases);
                bool nodeAlreadyInList = findNodeWithNameIn(currentPhase.name, phasesOfTheSkill) != null;
                bool nodeDoesNotExist = currentPhase == null;

                if (nodeAlreadyInList || nodeDoesNotExist) break;
            }

            return phasesOfTheSkill;
        }

        private PhaseNode findNodeWithNameIn(string phaseName, List<PhaseNode> whereToLookIn)
        {
            foreach (PhaseNode node in whereToLookIn)
                if (phaseName.Equals(node.name))
                    return node;
            return null;
        }
    }
}
