﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using RobotBehaviourEditor.Model;
using System.IO;
using System.Windows.Forms;

namespace RobotBehaviourEditor.Formats
{

    /**
     * A class rensponsible for parsing a List of JointMotions out
     * of an XML in a Jim format
     * 
     * */
    public class XmlImporterForJim : RobotBehaviourEditor.Formats.IXmlImporter
    {

        private string filePath;

        private List<PhaseNodeForJim> allPhases = new List<PhaseNodeForJim>();

        private string initialPhase = null;

        private string lowLevelSkill = null;

        private Dictionary<string, double> hashMapOfJoints = new Dictionary<string, double>();

        private string[] JOINT_NAMES = {"he1","he2","rle1","rle2","rle3","rle4","rle5","rle6","rae1","rae2","rae3",
                                            "rae4","lle1","lle2","lle3","lle4","lle5","lle6","lae1","lae2","lae3","lae4"};

        public XmlImporterForJim(string filePath)
        {
            this.filePath = filePath;
        }

        public List<string> lowLevelSkillsInFile()
        {
            XmlReader reader = createXmlReader();
            List<string> lowSkillNames = lowSkills(reader);

            if (lowSkillNames.Count == 0)
                throw new Exception("V súbore sa nenašiel žiaden low_skill tag");

            reader.Close();
            return lowSkillNames;
        }

        private XmlReader createXmlReader()
        {
            FileInfo xmlFile = new FileInfo(filePath);

            XmlReader reader = new XmlTextReader("file://" + xmlFile.FullName);
            return reader;
        }

        private List<string> lowSkills(XmlReader reader)
        {
            List<string> lowSkillNames = new List<string>();

            while (reader.Read())
                if (reader.NodeType == XmlNodeType.Element && reader.Name.Equals("low_skill"))
                {
                    while (reader.MoveToNextAttribute())
                        if ("name".Equals(reader.Name))
                            lowSkillNames.Add(reader.Value);
                }
            return lowSkillNames;
        }

        public List<JointMotion> movesOf(string lowLevelSkill)
        {
            XmlReader reader = createXmlReader();
            this.lowLevelSkill = lowLevelSkill;
            PhaseNodeForJim phase = null;

            while (reader.Read())
            {
                if (reader.NodeType == XmlNodeType.EndElement)
                    continue;
                if ("low_skill".Equals(reader.Name))
                    findInitialPhase(reader);

                if (reader.Name.Equals("phase"))
                {
                    phase = PhaseNodeForJim.parse(reader);
                    allPhases.Add(phase);
                }
                if (reader.Name.Equals("duration"))//duration
                    phase.speedConstant = parseSpeedConstant(reader);

                if (JOINT_NAMES.Contains(reader.Name))//effectors
                    phase.addMove(EffectorNodeForJim.parse(reader));
            }
            List<JointMotion> motions = new List<JointMotion>();
            double elapsedTime = 0.0;

            foreach (PhaseNodeForJim aPhase in phasesOfChosenSkill())
            {
                double durationOfThisPhase = aPhase.speedConstant;
                foreach (EffectorNodeForJim move in aPhase.getMoves())
                {
                    JointMotion motion = new JointMotion();
                    motion.Joint = Joints.GetJointTypeById(move.jointName);
                    motion.StartPosition = startAngleOfJoin(move.jointName, move.endAngle);
                    motion.EndPosition = move.endAngle;
                    motion.StartTime = (int)elapsedTime;
                    motion.Duration = (int)durationOfThisPhase;
                    motions.Add(motion);
                }
                elapsedTime = elapsedTime + durationOfThisPhase;
            }
            
            //merge motions same type into motion
            foreach (JointMotion outerMotion in motions)
            {
                foreach (JointMotion innerMotion in motions)
                {
                    if (outerMotion.IsActive && innerMotion.IsActive)
                    {
                        if (outerMotion != innerMotion)
                        {
                            if (outerMotion.Joint == innerMotion.Joint)
                            {
                                if (outerMotion.EndPosition == innerMotion.StartPosition)
                                {
                                    if (outerMotion.EndTime == innerMotion.StartTime)
                                    {
                                        outerMotion.EndPosition = innerMotion.EndPosition;
                                        outerMotion.Duration += innerMotion.Duration;
                                        innerMotion.IsActive = false;
                                    }
                                }
                            }
                        }
                    }
                }
            }

            //select active motion
            List<JointMotion> finalMotions = new List<JointMotion>();
            foreach (JointMotion motion in motions)
            {
                if (motion.IsActive)
                {
                    finalMotions.Add(motion);
                }
            }
            
            return finalMotions;
        }

        private void findInitialPhase(XmlReader reader)
        {
            while (reader.MoveToNextAttribute())
            {
                if ("firstPhase".Equals(reader.Name))
                {
                    initialPhase = reader.Value;
                }
            }
        }

        private int parseSpeedConstant(XmlReader reader)
        {
            reader.Read();
            return int.Parse(reader.Value);
        }

        private List<PhaseNodeForJim> phasesOfChosenSkill()
        {
            List<PhaseNodeForJim> phasesOfTheSkill = new List<PhaseNodeForJim>();
            PhaseNodeForJim currentPhase = findNodeWithNameIn(initialPhase, allPhases);

            while (!currentPhase.isFinal)
            {
                phasesOfTheSkill.Add(currentPhase);
                currentPhase = findNodeWithNameIn(currentPhase.nextName, allPhases);
                bool nodeAlreadyInList = findNodeWithNameIn(currentPhase.name, phasesOfTheSkill) != null;
                bool nodeDoesNotExist = currentPhase == null;

                if (nodeAlreadyInList || nodeDoesNotExist) break;
            }
            
            phasesOfTheSkill.Add(currentPhase);

            return phasesOfTheSkill;
        }

        private PhaseNodeForJim findNodeWithNameIn(string phaseName, List<PhaseNodeForJim> whereToLookIn)
        {
            foreach (PhaseNodeForJim node in whereToLookIn)
                if (phaseName.Equals(node.name))
                    return node;
            return null;
        }

        private double startAngleOfJoin(string joint, double endAngle)
        {
            double startAngle;
            //zoberieme poslednu hodnotu pre dany klb, co bude zaciatocna hodnota,
            //ak sa tam klb este nenachadza pociatocna hodnota je nula
            if (!hashMapOfJoints.TryGetValue(joint, out startAngle))
            {
                startAngle = 0.0;
            }

            //pridame novu hodnotu do hashMapy
            if(hashMapOfJoints.ContainsKey(joint))
            {
                hashMapOfJoints[joint] = endAngle;
            }
            else //ak sa tam klb uz raz nachadza
            {
                hashMapOfJoints.Add(joint, endAngle);
            }

            return startAngle;
        }
    }
}
