﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using RobotBehaviourEditor.MotionEditor.Visualisation;
using RobotBehaviourEditor.Model;
using System.Windows.Media.Media3D;

namespace RobotBehaviourEditor.Model.Physics
{
    class Body
    {
        List<BodyPart> bodyParts = new List<BodyPart>(Enum.GetValues(typeof(JointTypes)).Length);

        public Body(Robot3D robot, Gyroscope gyroscope)
        {
            InitializeBodyParts();
            SetJointAngles(robot);
            ComputeLocations(gyroscope);
        }

        void InitializeBodyParts()
        {
            BodyPart torso = new BodyPart(JointTypes.Torso, 0);

            BodyPart neck = new BodyPart(
                JointTypes.Neck,
                torso,
                InitialConfiguration.NeckTranslation,
                InitialConfiguration.NeckMass,
                InitialConfiguration.NeckAnchor,
                InitialConfiguration.NeckAxis);

            BodyPart head = new BodyPart(
                JointTypes.Head,
                neck,
                InitialConfiguration.HeadTranslation,
                InitialConfiguration.HeadMass,
                InitialConfiguration.HeadAnchor,
                InitialConfiguration.HeadAxis);

            bodyParts.Add(torso);
            bodyParts.Add(head);
            bodyParts.Add(neck);

            InitializeArmParts(torso, true);
            InitializeArmParts(torso, false);
            InitializeLegParts(torso, true);
            InitializeLegParts(torso, false);
        }

        void InitializeLegParts(BodyPart torso, bool isRight)
        {
            BodyPart hip1 = new BodyPart(
                isRight ? JointTypes.RightHip1 : JointTypes.LeftHip1,
                torso,
                isRight ? InitialConfiguration.RightHip1Translation : InitialConfiguration.LeftHip1Translation,
                InitialConfiguration.Hip1Mass,
                InitialConfiguration.Hip1Anchor,
                isRight ? InitialConfiguration.RightHip1Axis : InitialConfiguration.LeftHip1Axis);

            BodyPart hip2 = new BodyPart(
                isRight ? JointTypes.RightHip2 : JointTypes.LeftHip2,
                hip1,
                InitialConfiguration.Hip2Translation,
                InitialConfiguration.Hip2Mass,
                InitialConfiguration.Hip2Anchor,
                InitialConfiguration.Hip2Axis);

            BodyPart thigh = new BodyPart(
                isRight ? JointTypes.RightThigh : JointTypes.LeftThigh,
                hip2,
                InitialConfiguration.ThighTranslation,
                InitialConfiguration.ThighMass,
                InitialConfiguration.ThighAnchor,
                InitialConfiguration.ThighAxis);

            BodyPart shank = new BodyPart(
                isRight ? JointTypes.RightShank : JointTypes.LeftShank,
                thigh,
                InitialConfiguration.ShankTranslation,
                InitialConfiguration.ShankMass,
                InitialConfiguration.ShankAnchor,
                InitialConfiguration.ShankAxis);

            BodyPart ankle = new BodyPart(
                isRight ? JointTypes.RightAnkle : JointTypes.LeftAnkle,
                shank,
                InitialConfiguration.AnkleTranslation,
                InitialConfiguration.AnkleMass,
                InitialConfiguration.AnkleAnchor,
                InitialConfiguration.AnkleAxis);

            BodyPart foot = new BodyPart(
                isRight ? JointTypes.RightFoot : JointTypes.LeftFoot,
                ankle,
                InitialConfiguration.FootTranslation,
                InitialConfiguration.FootMass,
                InitialConfiguration.FootAnchor,
                InitialConfiguration.FootAxis);

            bodyParts.Add(hip1);
            bodyParts.Add(hip2);
            bodyParts.Add(thigh);
            bodyParts.Add(shank);
            bodyParts.Add(ankle);
            bodyParts.Add(foot);
        }

        void InitializeArmParts(BodyPart torso, bool isRight)
        {
            BodyPart shoulder = new BodyPart(
                isRight ? JointTypes.RightShoulder : JointTypes.LeftShoulder,
                torso,
                isRight ? InitialConfiguration.RightShoulderTranslation : InitialConfiguration.LeftShoulderTranslation,
                InitialConfiguration.ShoulderMass,
                InitialConfiguration.ShoulderAnchor,
                InitialConfiguration.ShoulderAxis);

            BodyPart upperarm = new BodyPart(
                isRight ? JointTypes.RightUpperArm : JointTypes.LeftUpperArm,
                shoulder,
                isRight ? InitialConfiguration.RightUpperarmTranslation : InitialConfiguration.LeftUpperarmTranslation,
                InitialConfiguration.ShoulderMass,
                InitialConfiguration.ShoulderAnchor,
                InitialConfiguration.ShoulderAxis);

            BodyPart elbow = new BodyPart(
                isRight ? JointTypes.RightElbow : JointTypes.LeftElbow,
                upperarm,
                isRight ? InitialConfiguration.RightElbowTranslation : InitialConfiguration.LeftElbowTranslation,
                InitialConfiguration.ElbowMass,
                InitialConfiguration.ElbowAnchor,
                InitialConfiguration.ElbowAxis);

            BodyPart lowerarm = new FrontalBodyPart(
                isRight ? JointTypes.RightLowerArm : JointTypes.LeftLowerArm,
                elbow,
                InitialConfiguration.LowerarmTranslation,
                InitialConfiguration.LowerarmMass,
                InitialConfiguration.LowerarmAnchor,
                InitialConfiguration.LowerarmAxis,
                InitialConfiguration.LowerarmFrontalPoints);

            bodyParts.Add(shoulder);
            bodyParts.Add(upperarm);
            bodyParts.Add(elbow);
            bodyParts.Add(lowerarm);
        }

        void SetJointAngles(Robot3D robot)
        {
            foreach (var part in bodyParts)
                part.JointAngle = robot.GetPartWithId(
                    (int)part.Name).CurrentAngle;
        }

        void ComputeLocations(Gyroscope gyroscope)
        {
            foreach (var part in bodyParts)
                part.ComputeLocation(gyroscope);
        }

        public List<Point3D> GetAllPoints()
        {
            List<Point3D> points = new List<Point3D>(bodyParts.Count * 3 / 2);

            foreach (var item in bodyParts)
            {
                points.Add(item.Location);

                if (item is FrontalBodyPart)
                    points.AddRange(((FrontalBodyPart)item).FrontalPoints);
            }

            return points;
        }

        /// <summary>
        /// tazisko
        /// hodnota sa vypocitava
        /// </summary>
        public Point3D CenterOfMass
        {
            get
            {
                //sumy sucinov hmotnosti a pozicii/suradnic
                double sumX = 0, sumY = 0, sumZ = 0;

                foreach (var part in bodyParts)
                {
                    sumX += part.Mass * part.Location.X;
                    sumY += part.Mass * part.Location.Y;
                    sumZ += part.Mass * part.Location.Z;
                }

                double totalMass = TotalMass;
                return new Point3D(
                    sumX / totalMass,
                    sumY / totalMass,
                    sumZ / totalMass);
            }
        }

        //hmotnost celeho hraca
        //hodnota sa vypocitava
        double TotalMass
        {
            get
            {
                double sum = 0;
                foreach (var part in bodyParts)
                    sum += part.Mass;

                return sum;
            }
        }
    }
}
