﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Media.Media3D;
using RobotBehaviourEditor.Model;

namespace RobotBehaviourEditor.Model.Physics
{
    class BodyPart
    {
        /// <summary>
        /// rodicovska cast tela, na ktorej je naviazana
        /// </summary>
        public BodyPart Parent { get; private set; }

        /// <summary>
        /// meno casti tela agenta
        /// hodnota z konfiguraku
        /// </summary>
        public JointTypes Name { get; private set; }

        /// <summary>
        /// vektor, posunutie od rodicovskej casti (v zaciatocnej konfiguracii)
        /// hodnota z konfiguraku
        /// </summary>
        public Vector3D Translation { get; private set; }

        /// <summary>
        /// hmotnost
        /// hodnota z konfiguraku
        /// </summary>
        public double Mass { get; private set; }

        /// <summary>
        /// miesto umiestnenia klbu
        /// hodnota z konfiguraku
        /// </summary>
        public Vector3D JointAnchor { get; private set; }

        /// <summary>
        /// vektor (os) otacania klbu
        /// hodnota z konfiguraku
        /// </summary>
        public Vector3D JointAxis { get; private set; }

        /// <summary>
        /// natocenie klbu
        /// posiela server, perceptor
        /// </summary>
        public double JointAngle { get; set; }

        /// <summary>
        /// pozicia casti tela vzhladom na torzo
        /// hodnotu vypocitavame
        /// </summary>
        public Point3D Location { get; private set; }
	
	    public BodyPart(JointTypes name, double mass)
        {
            this.Name = name;
            this.Mass = mass;
        }

	    public BodyPart(JointTypes name, BodyPart parent, Vector3D translation, double mass, Vector3D jointAnchor, Vector3D jointAxis)
        {
            this.Name = name;
            this.Parent = parent;
            this.Translation = translation;
            this.Mass = mass;
            this.JointAnchor = jointAnchor;
            this.JointAxis = jointAxis;
        }

        /// <summary>
        /// prepocita suradnice vstupneho budu vzhladom na rozmery a otocenie
        /// danej casti tela, novym zaciatocnym bodom suradnej sustavy bude bod
        /// na rodicovksej casti tela
        /// </summary>
        Point3D Transform(Point3D point)
        {
            //ak nie je parent, sme uz v torze -> nie je co riesit
            if (Parent == null)
                return point;

            //toto sa snazime dostat na 0,0,0
            Point3D parentLocation = Translation.GetInverted().GetDestinationPoint();

            //posunieme vsetko do klbu, stredu otacania
            point = point.GetTranslated(JointAnchor.GetInverted());
            parentLocation = parentLocation.GetTranslated(JointAnchor.GetInverted());

            //vsetko otocime podla natocenia klbu a vektoru (osi) otacania v klbe
            point = point.GetRotated(JointAxis, JointAngle);
            parentLocation = parentLocation.GetRotated(JointAxis, JointAngle);

            //nastavime rodicovku cast do 0,0,0
            point = point.GetTranslated(parentLocation.GetVector().GetInverted());

            //rekurzivne prepocitame bod vzhladom na rodicovsku cast
            return Parent.Transform(point);
        }

        /// <summary>
        /// kompletny prepocet bodu vzhladom na zem
        /// </summary>
        protected Point3D TransformAll(Point3D point, Gyroscope gyroscope)
        {
            //vypocet pozicie vzhladom na torzo
            point = Transform(point);

            //prepocet vzhladom na zem
            point = point.GetRotated(new Vector3D(1, 0, 0), gyroscope.Alpha);
            point = point.GetRotated(new Vector3D(0, 1, 0), gyroscope.Beta);
            point = point.GetRotated(new Vector3D(0, 0, 1), gyroscope.Gama);

            return point;
        }

        /// <summary>
        /// dopocita dalsie hodnoty
        /// hlavne poziciu casti tela najskor vzhladom na torzo, potom na zem
        /// </summary>
        public virtual void ComputeLocation(Gyroscope gyroscope)
        {
            //v pripade torza nema zmysel nieco pocitat
            if (Parent == null)
            {
                Location = new Point3D(0, 0, 0);
                return;
            }

            Location = TransformAll(new Point3D(0, 0, 0), gyroscope);
        }
    }
}
