﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Media.Media3D;

namespace RobotBehaviourEditor.Model.Physics
{
    /// <summary>
    /// pre vypocet konvexnej plochy je pouzity
    /// Gift wrapping algorithm akurat nepocitam pozicie
    /// bodov od priamky, ale uhly od osy x
    /// http://en.wikipedia.org/wiki/Gift_wrapping_algorithm
    /// </summary>
    class StabilityHull
    {
        const double StrongPointTolerance = 0.04;
        public List<Point> Points { get; private set; }
        public Point Center { get; private set; }

        public StabilityHull(Body body)
        {
            ChoosePoints(body);
            ComputeHull();
            ComputeCenter();
        }

        public StabilityHull(Point[] points)
        {
            this.Points = new List<Point>(points);
            ComputeHull();
            ComputeCenter();
        }

        /// <summary>
        /// vyberie body na tele hraca pre vypocet konvexnej plochy
        /// </summary>
        private void ChoosePoints(Body body)
        {   
            List<Point3D> allPoints = body.GetAllPoints();
            Points = new List<Point>(10);

            //vyberie najnizsiu z-ovu suradnicu z bodov
            double lowestZ = allPoints.Min(x => x.Z);

            foreach (var item in allPoints)
                if (item.Z < lowestZ + StrongPointTolerance)
                    Points.Add(new Point(item.X, item.Y));
        }

        private void ComputeHull()
        {
            HashSet<Point> hull = new HashSet<Point>();
            HashSet<Point> notDonePoints = new HashSet<Point>(Points);
            
            //vektor smerujuci z aktualneho bodu do predchadzajuceho
            Vector toPrevious = new Vector(0, -1);
            Point
                first = SelectLeftmost(notDonePoints),
                actual = first,
                next = SelectNext(notDonePoints, actual, toPrevious);;

            hull.Add(first);
            notDonePoints.Remove(first);

            while (IsHullIncomplete(toPrevious, first, actual, next))
            {
                hull.Add(next);
                notDonePoints.Remove(next);
                
                if (notDonePoints.Count == 0)
                    break;

                toPrevious = CreateVector(next, actual);
                actual = next;
                next = SelectNext(notDonePoints, actual, toPrevious);
            }

            Points = new List<Point>(hull);
        }

        private bool IsHullIncomplete(Vector toPrevious, Point first, Point actual, Point next)
        {
            if (first == actual)
                return true;

            Vector
                toFirst = CreateVector(actual, first),
                toNext = CreateVector(actual, next);

            double
                angleFirst = toPrevious.GetClockwiseAngleTo(toFirst),
                angleNext = toPrevious.GetClockwiseAngleTo(toNext);

            //ak nahodou prvych viac ako 2 bodov su na priamke
            if (angleFirst == 0)
                return true;

            return angleFirst > angleNext;
        }

        private static Vector CreateVector(Point from, Point to)
        {
            return new Vector(to.X - from.X, to.Y - from.Y);
        }

        /// <summary>
        /// vyberie dalsi bod z mnoziny bodov tak,
        /// aby uhol vektoru z aktualneho bodu do neho
        /// v smere hodinovych ruciciek bol vzhladom
        /// na zadany vektor cim mensi
        /// </summary>
        private static Point SelectNext(HashSet<Point> notDonePoints, Point actual, Vector toPrevious)
        {
            double bestAngle = 2 * Math.PI;
            Point best = notDonePoints.First();

            foreach (var item in notDonePoints)
            {
                //aktualny bod nemoze byt aj dalsi
                //iba v pripade prveho bodu
                if (item == actual)
                    continue;
                
                Vector actualToItem = CreateVector(actual, item);
                double itemAngle = toPrevious.GetClockwiseAngleTo(actualToItem);

                if (itemAngle < bestAngle)
                {
                    best = item;
                    bestAngle = itemAngle;
                }
            }

            return best;
        }

        /// <summary>
        /// vyberie najlavejsi bod, ak ich je viac, tak najspodnejsi
        /// </summary>
        private static Point SelectLeftmost(HashSet<Point> points)
        {
            Point best = new Point(Double.MaxValue, 0);

            foreach (var item in points)
                if (item.X < best.X ||
                    (item.X == best.X && item.Y < best.Y))
                    best = item;

            return best;
        }

        private void ComputeCenter()
        {
            double sumX = 0, sumY = 0;

            foreach (var item in Points)
            {
                sumX += item.X;
                sumY += item.Y;
            }

            Center = new Point(sumX / Points.Count, sumY / Points.Count);
        }
    }
}
