﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using RobotBehaviourEditor.MotionEditor;

namespace RobotBehaviourEditor.Model
{
    public class RobotMotion
    {
        private string name = "";
        public string Name
        {
            get { return name; }
            set { name = value; }
        }

        private string description = "";
        public string Description
        {
            get { return description; }
            set { description = value; }
        }

        private List<State> states = new List<State>();
        public List<State> States
        {
            get { return states; }            
        }

        public RobotMotion()
        {
        }

        public RobotMotion(RobotMotion motion)
        {
            this.description = motion.description;
            this.name = motion.name;
            this.states = new List<State>();
            for (int i = 0; i < motion.states.Count; i++)
            {
                this.states.Add(new State(motion.states[i]));
            }            
        }

        public void AddState(State state)
        {
            states.Add(state);
        }

        public void Save(string fileName)
        {
            StreamWriter sw = new StreamWriter(fileName, false, Encoding.UTF8);
            sw.WriteLine("[MOTION]");
            sw.WriteLine("NAME=" + name);
            sw.WriteLine("DESCRIPTION=" + description);
            
            for (int i = 0; i < this.States.Count; i++)
            {
                sw.WriteLine("[STATE]");
                sw.WriteLine("INDEX=" + i.ToString());
                sw.WriteLine("NAME=" + States[i].Name);
                sw.WriteLine("SYNCH=" + GetString(States[i].Synchronized));
                for (int j = 0; j < States[i].Motions.Count; j++)
                {                    
                    sw.WriteLine("[JOINTMOTION]");
                    sw.WriteLine("INDEX=" + (j + 1).ToString());
                    sw.WriteLine("JOINT=" + Joints.GetJointByType(States[i].Motions[j].Joint).Id.ToLower());
                    sw.WriteLine("TIME=" + States[i].Motions[j].StartTime);
                    sw.WriteLine("DUR=" + States[i].Motions[j].Duration);
                    sw.WriteLine("STARTPOS=" + States[i].Motions[j].StartPosition);
                    sw.WriteLine("ENDPOS=" + States[i].Motions[j].EndPosition);                    
                }
                //sw.WriteLine("NEXT=" + action.States[i].Next);
            }

            sw.Close();
        }

        public static string GetString(bool value)
        {
            if (value) return "true"; else return "false";
        }

        public static RobotMotion Load(string fileName)
        {
            RobotMotion newAction = new RobotMotion();
            State curState = null;
            JointMotion curMotion = null;

            StreamReader sr = new StreamReader(fileName, Encoding.UTF8, true);
            
            LoadingState state =  LoadingState.Start;
            string line;
            while ((line = sr.ReadLine()) != null)
            {
                line = line.Trim();

                switch (line)
                {
                    case "[MOTION]":
                        state = LoadingState.Action;
                        break;
                    case "[STATE]":
                        state = LoadingState.State;
                        curState = new State();
                        newAction.AddState(curState);
                        break;
                    case "[JOINTMOTION]":
                        state = LoadingState.Motion;
                        curMotion = new JointMotion();
                        curState.AddMotion(curMotion);
                        break;
                }

                int valueStart = line.IndexOf('=') + 1;
                string value = "";
                if (valueStart > 0) value = line.Substring(valueStart).Trim();
                switch (state)
                {                    
                    case LoadingState.Action:
                        if (line.StartsWith("NAME="))
                            newAction.Name = value;
                        if (line.StartsWith("DESCRIPTION="))
                            newAction.Description = value;
                        break;
                    case LoadingState.State:
                        if (line.StartsWith("NAME="))
                            curState.Name = value;
                        if (line.StartsWith("SYNCH="))
                            curState.Synchronized = GetBool(value);
                        break;
                    case LoadingState.Motion:
                        if (line.StartsWith("JOINT="))
                            curMotion.Joint = Joints.GetJointTypeById(value);
                        if (line.StartsWith("TIME="))
                            curMotion.StartTime = int.Parse(value);
                        if (line.StartsWith("DUR="))
                            curMotion.Duration = int.Parse(value);
                        if (line.StartsWith("STARTPOS="))
                            curMotion.StartPosition = double.Parse(value);
                        if (line.StartsWith("ENDPOS="))
                            curMotion.EndPosition = double.Parse(value);
                        break;
                    default:
                        break;
                }

            }

            sr.Close();
            return newAction;
        }

        private static bool GetBool(string value)
        {
            if (value == "true") return true; else return false;
        }
        
        #region Loading States
        private enum LoadingState
        {
            Start = 0,
            Action = 1,
            State = 2,
            Motion = 3
        }
        #endregion

        public void ApplySettings(RunSettings settings)
        {
            bool firstState = true;
            foreach (State state in states)
            {                
                foreach (JointMotion motion in state.Motions)
                {
                    motion.Duration = (int)(motion.Duration / settings.Speed);
                    motion.StartTime = (int)(motion.StartTime / settings.Speed);

                    if (firstState)
                    {
                        //motion.Duration += settings.StartDelay;
                        motion.StartTime += settings.StartDelay;
                    }
                    else
                    {
                        //motion.Duration += settings.StateDelay;
                        motion.StartTime += settings.StateDelay;
                    }
                }
                firstState=false;
            }
        }
        
    }
}
