﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using RobotBehaviourEditor.Model;
using System.Drawing.Drawing2D;
using RobotBehaviourEditor.Utils;
using RobotBehaviourEditor.MotionEditor.Visualisation;
using System.Threading;

namespace RobotBehaviourEditor.MotionEditor
{
    public partial class MotionDiagram : UserControl
    {
        #region Fields
        Color[] colors = new Color[] { Color.Blue, Color.Red, Color.Green, Color.Purple, Color.Yellow, Color.Magenta };

        private Rectangle[] motionBounds;
        GripType selectedGrip = GripType.Middle;
        int selectedBar = -1;
        int startMouseX = -1;

        DateTime lastUpdateTime;

        private bool controlPressed = false;
        private bool shiftPressed = false;

        BackgroundWorker thumbWorker = new BackgroundWorker();
        private List<Bitmap> thumbnails = new List<Bitmap>();
        #endregion

        #region Properties
        [Browsable(false)]
        private List<JointMotion> jointMotions =  new List<JointMotion>();
        public IList<JointMotion> JointMotions
        {
            get { return jointMotions.AsReadOnly(); }            
        }

        private float msPerPixel = 10f;
        public float MsPerPixel
        {
            get { return msPerPixel; }
            set { msPerPixel = value; }
        }

        [Browsable(false)]
        private int hoverIndex = -1;
        public int HoverIndex
        {
            get { return hoverIndex; }
            set { hoverIndex = value; }
        }
        
        private int selectedIndex = -1;
        public int SelectedIndex
        {
            get { return selectedIndex; }
            set
            {
                selectedIndices.Clear();
                if (value != -1)
                    selectedIndices.Add(value);
                selectedIndex = value;
            }
        }
        
        [Browsable(false)]
        public JointMotion SelectedItem
        {
            get 
            {
                if (selectedIndex >= 0 && selectedIndex <= jointMotions.Count)
                    return jointMotions[selectedIndex];
                else
                    return null;
            }
        }

        private List<int> selectedIndices = new List<int>();
        [Browsable(false)]
        public int[] SelectedIndices
        {
            get
            {
                return selectedIndices.ToArray();
            }
            set
            {
                selectedIndices.Clear();
                for (int i = 0; i < value.Length; i++)
                {
                    if (value[i] != -1) selectedIndices.Add(value[i]);
                }
            }
        }

        private int rowHeight = 40;
        public int RowHeight
        {
            get { return rowHeight; }
            set { rowHeight = value; }
        }

        private int barHeight = 25;
        public int BarHeight
        {
            get { return barHeight; }
            set { barHeight = value; }
        }

        private Robot3DControl visualisationControl;
        public Robot3DControl VisualisationControl
        {
            get { return visualisationControl; }
            set { visualisationControl = value; }
        }
        #endregion

        public MotionDiagram()
        {
            InitializeComponent();
            
            thumbWorker.DoWork += new DoWorkEventHandler(thumbWorker_DoWork);
        }

        public void Clear()
        {
            selectedIndex = -1;
            jointMotions.Clear();
            this.Invalidate();
        }

        public void AddJoinMotion(JointMotion jointMotion)                
        {
            jointMotions.Add(jointMotion);
            UpdateScrollbarRanges();
            this.Invalidate();

            UpdateThumbnailsOnBackground();
        }        

        public void AddJoinMotions(JointMotion[] jointMotion)
        {
            jointMotions.AddRange(jointMotion);
            UpdateScrollbarRanges();
            this.Invalidate();

            UpdateThumbnailsOnBackground();
        }

        public void UpdateMotion(int index, JointMotion jointMotion)
        {
            jointMotions[index] = jointMotion;

            UpdateThumbnailsOnBackground();
        }

        public void RemoveAt(int index)
        {
            if (index >= 0)
            {
                jointMotions.RemoveAt(index);
                if (SelectedIndex > jointMotions.Count - 1) SelectedIndex--;

                UpdateScrollbarRanges();
                this.Invalidate();
            }
        }

        private void MotionDiagram_Paint(object sender, PaintEventArgs e)
        {
            Graphics g = e.Graphics;
            g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
            g.PixelOffsetMode = PixelOffsetMode.Default;
            g.CompositingQuality = CompositingQuality.HighQuality;            

            Rectangle boundsRect = new Rectangle(0, 0, this.Width - 1, this.Height - 1);
            g.FillRectangle(Brushes.White, boundsRect);
            g.DrawRectangle(Pens.Black, boundsRect);            

            if (jointMotions.Count > 0)
            {                
                motionBounds = new Rectangle[jointMotions.Count];

                //(int)moves.Max(x => x.StartTime + x.Duration);

                DrawGrid(g);

                // Pohyby
                StringFormat sf = new StringFormat() { LineAlignment = StringAlignment.Center };
                int y = -verticalScroll.Value;
                for (int i = 0; i < jointMotions.Count; i++)
                {
                    if (y + rowHeight > 0 && y < this.Height)
                    {
                        DrawJointMotion(g, i, y);
                    }
                    else if (y > this.Height)
                        break;

                    y += RowHeight;
                }

                DrawSelectedMotion(g);
            }

        }

        private void DrawJointMotion(Graphics g, int index, int y)
        {
            Brush textBrush = SystemBrushes.WindowText;
            Rectangle rowRect = new Rectangle(0, y, this.Width, RowHeight);

            if (HoverIndex == index)
            {
                g.FillRectangle(new SolidBrush(Color.FromArgb(70, SystemColors.Highlight)), rowRect);
            }

            if (SelectedIndices.Contains(index))
            {
                if (SelectedIndex == index)
                    GraphicsUtils.DrawHighlighting(g, rowRect);
                else
                    g.FillRectangle(new LinearGradientBrush(rowRect, SystemColors.Highlight, SystemColors.Highlight, LinearGradientMode.Vertical), rowRect);

                textBrush = SystemBrushes.HighlightText;
            }

            int x = (int)((double)(jointMotions[index].StartTime - horizontalScroll.Value) / msPerPixel);
            int w = Math.Max((int)((double)jointMotions[index].Duration / msPerPixel), 1);

            Rectangle rect = new Rectangle(x, y + 2, w, BarHeight - 4);
            motionBounds[index] = rect;

            g.FillRectangle(new LinearGradientBrush(rect, Color.White, Joints.GetJointByType(jointMotions[index].Joint).Color, LinearGradientMode.Vertical), rect);
            if (this.SelectedIndex == index) g.FillRectangle(new SolidBrush(Color.FromArgb(60, Color.White)), rect);
            g.DrawRectangle(Pens.Black, rect);

            if (SelectedIndices.Contains(index))
            {
                g.PixelOffsetMode = PixelOffsetMode.Half;
                Rectangle rect2 = new Rectangle(rect.X + 2, rect.Y + 2, rect.Width - 4, rect.Height - 4);
                g.DrawRectangle(new Pen(Color.Black, 1) { DashStyle = DashStyle.Dot }, rect2);
                g.PixelOffsetMode = PixelOffsetMode.Default;
            }

            Rectangle textRect = new Rectangle(0, y + barHeight, this.Width, rowHeight - barHeight);
            g.DrawString(JointMotions[index].ToString(), Font, textBrush, textRect);
        }

        private void DrawSelectedMotion(Graphics g)
        {
            // Oznaceny pohyb
            if (SelectedIndex >= 0 && SelectedIndex < jointMotions.Count)
            {
                int x = (int)((double)(jointMotions[SelectedIndex].StartTime - horizontalScroll.Value) / msPerPixel);
                int w = Math.Max((int)((double)jointMotions[SelectedIndex].Duration / msPerPixel), 1);
                Rectangle rect;

                if (x > 0)
                {
                    rect = new Rectangle(0, 0, x, this.Height);
                    g.FillRectangle(new LinearGradientBrush(rect, Color.Transparent, Color.FromArgb(80, Color.Black), LinearGradientMode.Horizontal), rect);
                }
                if (x < this.Width)
                {
                    rect = new Rectangle(x + w, 0, this.Width, this.Height);
                    g.FillRectangle(new LinearGradientBrush(rect, Color.FromArgb(80, Color.Black), Color.Transparent, LinearGradientMode.Horizontal), rect);
                }
            }
        }

        private void DrawGrid(Graphics g)
        {
            // Mriezka
            Pen dsecPen = new Pen(Color.FromArgb(30, Color.Black));
            Pen secPen = new Pen(Color.FromArgb(50, Color.Black));

            int thumbWidth = (int)(1000 / msPerPixel);
            int thumbHeight = thumbWidth;

            for (float i = 0; i < this.Width; i += 500 / msPerPixel)
            {
                int x = (int)(i - horizontalScroll.Value / msPerPixel);
                if (Math.Round(i * msPerPixel) % 1000 == 0)
                    g.DrawLine(secPen, x, 0, x, this.Height);
                else
                    g.DrawLine(dsecPen, x, 0, x, this.Height);

                if ((i * msPerPixel) % 1000 == 0)
                {
                    int ti = (int)(i / 1000);
                    if (ti < thumbnails.Count)
                    {
                        g.DrawImage(thumbnails[ti], new Rectangle(x, 0, thumbWidth, thumbHeight));
                    }
                }
            }
        }

        private int GetMotionBarAtPosition(int x, int y)
        {
            if (motionBounds == null)
            {
                return -1;
            }
            else
            {
                for (int i = 0; i < motionBounds.Length; i++)
                {
                    if (motionBounds[i].Contains(x, y))
                        return i;
                }
                return -1;
            }
        }

        enum GripType
        {
            Left = 0, Middle = 1, Right = 2
        }
        private GripType GetGripTypeAtPosition(int x, int y)
        {
            int index = GetMotionBarAtPosition(x, y);

            if (index >= 0)
            {
                if (Math.Abs(motionBounds[index].X - x) < 5)
                    return GripType.Left;
                else if (Math.Abs(motionBounds[index].Right - x) < 5)
                    return GripType.Right;
                else
                    return GripType.Middle;
            }
            return GripType.Middle;
        }

        private void MotionDiagram_MouseDown(object sender, MouseEventArgs e)
        {
            selectedBar = GetMotionBarAtPosition(e.X, e.Y);
            selectedGrip = GetGripTypeAtPosition(e.X, e.Y);
            startMouseX = e.X;

            if (!shiftPressed)
            {
                if (!selectedIndices.Contains(selectedBar))
                {
                    if (HoverIndex < 0 || HoverIndex >= JointMotions.Count)
                    {
                        if (!controlPressed)
                            SelectedIndex = -1;
                    }
                    else
                    {
                        if (controlPressed)
                        {
                            if (hoverIndex != -1)
                            {
                                if (selectedIndices.Contains(HoverIndex))
                                    selectedIndices.Remove(HoverIndex);
                                else
                                    selectedIndices.Add(HoverIndex);
                            }
                        }
                        else
                            SelectedIndex = HoverIndex;
                    }
                }
            }
            else
            {
                selectedIndices.Clear();
                int from = Math.Min(hoverIndex, selectedIndex);
                int to = Math.Max(hoverIndex, selectedIndex);
                for (int i = from; i <= to; i++)
                {
                    if (i != -1)
                    {
                        selectedIndices.Add(i);
                    }
                }
                //selectedIndex = hoverIndex;
            }

            this.Invalidate();            
        }

        private void MotionDiagram_MouseMove(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.None)
            {
                InvalidateHoverItem();
                HoverIndex = (int)Math.Floor(((double)e.Y + verticalScroll.Value) / RowHeight);
                InvalidateHoverItem();

                int hoverBar = GetMotionBarAtPosition(e.X, e.Y);
                if (hoverBar != -1)
                {
                    GripType grip = GetGripTypeAtPosition(e.X, e.Y);
                    switch (grip)
                    {
                        case GripType.Left:
                            this.Cursor = Cursors.SizeWE;
                            break;
                        case GripType.Middle:
                            this.Cursor = Cursors.SizeAll;
                            break;
                        case GripType.Right:
                            this.Cursor = Cursors.SizeWE;
                            break;
                    }
                }
                else
                {
                    this.Cursor = Cursors.Default;
                }
            }
            else if (e.Button == MouseButtons.Left)
            {
                if (selectedBar != -1)
                {
                    //InvalidateSelectedItem();

                    int dif = (int)((startMouseX - e.X) * msPerPixel);
                    startMouseX = e.X;

                    MoveSelectedMotionsInTime(selectedGrip, dif);

                    this.Invalidate();
                    //if ((DateTime.Now.Subtract(lastUpdateTime).TotalMilliseconds > 500))
                    //{
                    //    lastUpdateTime = DateTime.Now;
                    //    this.Invalidate();
                    //}                    
                }                                        
            }
        }

        private void InvalidateSelectedItem()
        {
            Rectangle rect =
                    new Rectangle(0, selectedBar * rowHeight - verticalScroll.Value,
                        this.Width, rowHeight);
            this.Invalidate(rect);
        }

        private void InvalidateHoverItem()
        {
            Rectangle rect =
                    new Rectangle(0, hoverIndex * rowHeight - verticalScroll.Value,
                        this.Width, rowHeight);
            this.Invalidate(rect);
        }

        private void MoveSelectedMotionsInTime(GripType gripType, int timeDifference)
        {
            foreach (int i in selectedIndices)
            {
                switch (selectedGrip)
                {
                    case GripType.Left:
                        jointMotions[i].StartTime -= timeDifference;
                        jointMotions[i].Duration += timeDifference;
                        break;
                    case GripType.Middle:
                        jointMotions[i].StartTime -= timeDifference;
                        break;
                    case GripType.Right:
                        jointMotions[i].Duration -= timeDifference;
                        break;
                }
            }

            UpdateThumbnailsOnBackground();
        }

        private void MotionDiagram_MouseLeave(object sender, EventArgs e)
        {

        }

        private void MotionDiagram_MouseUp(object sender, MouseEventArgs e)
        {            
            this.Invalidate();
        }

        private void MotionDiagram_Resize(object sender, EventArgs e)
        {
            UpdateScrollbarRanges();
            this.Invalidate();
        }

        private void UpdateScrollbarRanges()
        {
            verticalScroll.Maximum = jointMotions.Count * rowHeight;
        }
        
        protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
        {
            if (keyData == Keys.Down || keyData == Keys.Up)
            {
                MotionDiagram_KeyDown(this, new KeyEventArgs(keyData));
                return true;
            }

            return base.ProcessCmdKey(ref msg, keyData);
        }

        private void MotionDiagram_KeyDown(object sender, KeyEventArgs e)
        {
            controlPressed = e.Control;
            shiftPressed = e.Shift;

            switch (e.KeyCode)
            {
                case Keys.Up:
                    if (SelectedIndex > 0)
                    {
                        SelectedIndex--;
                        this.Invalidate();
                    }
                    break;
                case Keys.Down:
                    if (SelectedIndex < JointMotions.Count - 1)
                    {
                        SelectedIndex++;
                        this.Invalidate();
                    }
                    break;
                case Keys.Delete:
                    DeleteSelected();
                    break;
            }
        }

        private void verticalScroll_Scroll(object sender, ScrollEventArgs e)
        {
            this.Invalidate();
        }

        private void horizontalScroll_Scroll(object sender, ScrollEventArgs e)
        {
            this.Invalidate();
        }

        private void horizontalScroll_ValueChanged(object sender, EventArgs e)
        {
            //horizontalScroll_Scroll(sender, null);
        }

        private void verticalScroll_ValueChanged(object sender, EventArgs e)
        {
            //verticalScroll_Scroll(sender, null);
        }

        private void MotionDiagram_KeyUp(object sender, KeyEventArgs e)
        {
            controlPressed = e.Control;
            shiftPressed = e.Shift;
        }

        public void UpdateThumbnailsOnBackground()
        {
            //if (!thumbWorker.IsBusy)
            //{                
            //    thumbWorker.RunWorkerAsync();                
            //}
        }

        void thumbWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            if (visualisationControl != null)
            {
                JointMotion lastMotion = new RobotAnimation(jointMotions).GetLastMotion();
                int endTime = lastMotion.StartTime + lastMotion.Duration;
                List<Bitmap> newThumbs = new List<Bitmap>();

                for (int i = 0; i < endTime; i += 1000)
                {
                    newThumbs.Add(visualisationControl.GetRobotMotionPreview(i));
                }

                DisposeThumbnails();
                thumbnails = newThumbs;
            }
        }        

        public void UpdateThumbnails()
        {
            JointMotion lastMotion = new RobotAnimation(jointMotions).GetLastMotion();
            int endTime = lastMotion.StartTime + lastMotion.Duration;

            DisposeThumbnails();

            for (int i = 0; i < endTime; i += 1000)
            {
                thumbnails.Add(visualisationControl.GetRobotMotionPreview(i));
            }
        }

        private void DisposeThumbnails()
        {
            foreach (var t in thumbnails)
                t.Dispose();
            thumbnails.Clear();
        }

        protected override void OnHandleDestroyed(EventArgs e)
        {
            DisposeThumbnails();
            base.OnHandleDestroyed(e);
        }

        public void DeleteSelected()
        {
            var indices = this.SelectedIndices.OrderByDescending(x => x);
            foreach (var i in indices)
            {
                this.RemoveAt(i);
            }
            this.SelectedIndex = -1;
        }

        public void MoveUpSelected()
        {
            if (this.SelectedIndices.Length != 0)
            {
                var indices = this.SelectedIndices.OrderBy(x => x).ToList();
                for (int i = 0; i < jointMotions.Count; i++)
                {
                    if (indices.Contains(i) && i > indices.IndexOf(i))
                    {
                        JointMotion tmp = jointMotions[i - 1];
                        jointMotions[i - 1] = jointMotions[i];
                        jointMotions[i] = tmp;
                        indices[indices.IndexOf(i)]--;
                    }
                }
                this.SelectedIndices = indices.ToArray();
                /*if (indices[0] != 0)
                {
                    this.SelectedIndices = indices.Select(x => x - 1).ToArray();
                }*/
            }
        }

        public void MoveDownSelected()
        {
            if (this.SelectedIndices.Length != 0)
            {
                var indices = this.SelectedIndices.OrderByDescending(x => x).ToList();
                for (int i = jointMotions.Count - 1; i >= 0; i--)
                {
                    if (indices.Contains(i) && i < jointMotions.Count - 1 - indices.IndexOf(i))
                    {
                        JointMotion tmp = jointMotions[i + 1];
                        jointMotions[i + 1] = jointMotions[i];
                        jointMotions[i] = tmp;
                        indices[indices.IndexOf(i)]++;
                    }
                }
                this.SelectedIndices = indices.ToArray();
                /*if (indices[indices.Count - 1] != indices.Count - 1)
                {
                    this.SelectedIndices = indices.Select(x => x + 1).ToArray();
                }*/
            }
        }
    }
}
