﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Windows.Media.Media3D;
using System.Windows.Media.Animation;
using System.Timers;
using System.ComponentModel;
using System.Windows.Threading;
using RobotBehaviourEditor.Model;
using System.IO;
using RobotBehaviourEditor.Model.Physics;

namespace RobotBehaviourEditor.MotionEditor.Visualisation
{
    /// <summary>
    /// Interaction logic for UserControl1.xaml
    /// </summary>
    public partial class Robot3DControl : UserControl
    {
        private NamedModel3D selectedModel = null;
        private NamedModel3D hoverModel = null;

        private Material hoverMaterial = new DiffuseMaterial(new ImageBrush(Images.HoverTexture));
        private Material selectedMaterial = new DiffuseMaterial(new ImageBrush(Images.SelectedTexture));

        private Point previousPosition2D;
        private Vector3D previousPosition3D = new Vector3D(0, 0, 1);
        private bool trackStart = false;

        private Transform3DGroup transform;
        private ScaleTransform3D scale = new ScaleTransform3D();
        private AxisAngleRotation3D rotation = new AxisAngleRotation3D();

        private bool slidersInitilization = false;

        Robot3D robot3d = null;
        Robot3DPart selectedPart = null;        
        
        RobotAnimator animator = null;

        private List<JointMotion> jointMotions = new List<JointMotion>();
        public IList<JointMotion> JointMotions
        {
            get
            {
                return jointMotions;
            }
            set
            {
                this.jointMotions = new List<JointMotion>(value);

                JointMotion lastMotion = new RobotAnimation(jointMotions).GetLastMotion();
                if (lastMotion !=null)               
                    AnimationSlider.Maximum = (lastMotion.EndTime);                

                UpdateStepList();
            }
        }

#region EventHandling
        private FrameworkElement eventSource;
        public FrameworkElement EventSource
        {
            get { return eventSource; }

            set
            {
                if (eventSource != null)
                {
                    eventSource.MouseDown -= this.OnMouseDown;
                    eventSource.MouseUp -= this.OnMouseUp;
                    eventSource.MouseMove -= this.OnMouseMove;                    
                }

                eventSource = value;

                eventSource.MouseDown += this.OnMouseDown;
                eventSource.MouseUp += this.OnMouseUp;
                eventSource.MouseMove += this.OnMouseMove;                
            }
        }        
#endregion

        public Robot3DControl()
        {
            InitializeComponent();
            Initialize();
        }

        public Robot3DControl(Window window)
        {
            InitializeComponent();
            Initialize();

            window.KeyDown += new KeyEventHandler(OnKeyDown);
        }
        
        private void Initialize()
        {
            transform = new Transform3DGroup();
            transform.Children.Add(scale);
            transform.Children.Add(new RotateTransform3D(rotation));

            EventSource = CaptureBorder;

            robot3d = Robot3D.GetRobot();
            robot3d.InsertToViewport(robotViewport);
            scale.ScaleX *= robot3d.DefaultScale;
            scale.ScaleY *= robot3d.DefaultScale;
            scale.ScaleZ *= robot3d.DefaultScale;

            robotViewport.Camera.Transform = transform; 

            if (Program.Config.VisualisationCenterOfMass)
            {
                robot3d.CenterOfMassCuboid = new Cuboid(new Rect3D(new Point3D(0, 0, 0), new Size3D(0.1, 0.1, 0.1)),
                     new ImageBrush(Images.ComTexture));                
                robot3d.CenterOfMassCuboid.InsertToViewport(robotViewport, "CenterOfMass", -1);
            }

            animator = new RobotAnimator(robot3d);
            
            //ModelVisual3D model = new ModelVisual3D();
            //model.Content = robot;
            //robotViewport.Children.Add(model);
        }        

        private void Track(Point currentPosition)
        {
            Vector3D currentPosition3D = ProjectToTrackball(
                EventSource.ActualWidth, EventSource.ActualHeight, currentPosition);

            Vector3D axis = Vector3D.CrossProduct(previousPosition3D, currentPosition3D);
            double angle = Vector3D.AngleBetween(previousPosition3D, currentPosition3D);

            if (angle != 0)
            {
                Quaternion delta = new Quaternion(axis, -angle);

                // Get the current orientantion from the RotateTransform3D
                AxisAngleRotation3D r = rotation;
                Quaternion q = new Quaternion(rotation.Axis, rotation.Angle);

                // Compose the delta with the previous orientation
                q *= delta;

                // Write the new orientation back to the Rotation3D
                rotation.Axis = q.Axis;
                rotation.Angle = q.Angle;

                previousPosition3D = currentPosition3D;
            }
        }

        private Vector3D ProjectToTrackball(double width, double height, Point point)
        {
            double x = point.X / (width / 2);    // Scale so bounds map to [0,0] - [2,2]
            double y = point.Y / (height / 2);

            x = x - 1;                           // Translate 0,0 to the center
            y = 1 - y;                           // Flip so +Y is up instead of down

            double z2 = 1 - x * x - y * y;       // z^2 = 1 - x^2 - y^2
            double z = z2 > 0 ? Math.Sqrt(z2) : 0;

            return new Vector3D(x, y, z);
        }

        private void Zoom(Point currentPosition)
        {
            double yDelta = currentPosition.Y - previousPosition2D.Y;

            double scale = Math.Exp(yDelta / 100);    // e^(yDelta/100) is fairly arbitrary.

            this.scale.ScaleX *= scale;
            this.scale.ScaleY *= scale;
            this.scale.ScaleZ *= scale;
        }

        private void OnMouseDown(object sender, MouseEventArgs e)
        {            
            SelectRobotPart(e.GetPosition(robotViewport));            

            Mouse.Capture(EventSource, CaptureMode.Element);
            previousPosition2D = e.GetPosition(EventSource);
            previousPosition3D = ProjectToTrackball(
                EventSource.ActualWidth,
                EventSource.ActualHeight,
                previousPosition2D);
            trackStart = true;
        }        

        private void OnMouseUp(object sender, MouseEventArgs e)
        {
            Mouse.Capture(EventSource, CaptureMode.None);
            trackStart = false;
        }

        private void OnMouseMove(object sender, MouseEventArgs e)
        {
            HitTestResult result = VisualTreeHelper.HitTest(robotViewport, e.GetPosition(robotViewport));
            if (result != null)
            {
                NamedModel3D hitModel = result.VisualHit as NamedModel3D;

                if (hitModel.Id != -1)
                {
                    if (hitModel != null && hitModel != hoverModel)
                    {
                        if (hoverModel != null) hoverModel.RestoreMaterial();
                        hitModel.SaveMaterial();
                        hitModel.ChangeMaterial(hoverMaterial);
                    }
                    hoverModel = hitModel;
                }
            }

            if (trackStart)
            {
                Point currentPosition = e.GetPosition(EventSource);

                // Prefer tracking to zooming if both buttons are pressed.
                if (e.LeftButton == MouseButtonState.Pressed)
                {
                    Track(currentPosition);
                }
                else if (e.RightButton == MouseButtonState.Pressed)
                {
                    Zoom(currentPosition);
                }

                previousPosition2D = currentPosition;

                robotViewport.Camera.Transform = transform;
            }
        }

        private void OnKeyDown(object sender, KeyEventArgs e)
        {
            Point3D pos = ((PerspectiveCamera)robotViewport.Camera).Position;

            switch (e.Key)
            {
                case Key.Left:
                    ((PerspectiveCamera)robotViewport.Camera).Position =
                        new Point3D(pos.X - 5, pos.Y, pos.Z);
                    break;
                case Key.Right:
                    ((PerspectiveCamera)robotViewport.Camera).Position =
                        new Point3D(pos.X + 5, pos.Y, pos.Z);
                    break;
                case Key.Down:
                    ((PerspectiveCamera)robotViewport.Camera).Position =
                        new Point3D(pos.X, pos.Y, pos.Z - 5);
                    break;
                case Key.Up:
                    ((PerspectiveCamera)robotViewport.Camera).Position =
                        new Point3D(pos.X, pos.Y, pos.Z + 5);
                    break;
                case Key.PageDown:
                    ((PerspectiveCamera)robotViewport.Camera).Position =
                        new Point3D(pos.X, pos.Y - 5, pos.Z);
                    break;
                case Key.PageUp:
                    ((PerspectiveCamera)robotViewport.Camera).Position =
                        new Point3D(pos.X, pos.Y + 5, pos.Z);
                    break;
                case Key.Space:
                    RotateRobot();
                    break;
            }
        }

        private bool SelectRobotPart(Point point)
        {
            HitTestResult result = VisualTreeHelper.HitTest(robotViewport, point);
            if (result != null)
            {
                NamedModel3D hitModel = result.VisualHit as NamedModel3D;
                if (hitModel.Id != -1)
                {
                    if (hitModel != null || hitModel != selectedModel)
                    {
                        hoverModel = null;
                        if (selectedModel != null) selectedModel.SetOriginalMaterial();
                        hitModel.ChangeMaterial(selectedMaterial);
                        selectedModel = hitModel;
                        selectedPart = robot3d.GetPartWithId(selectedModel.Id);

                        slidersInitilization = true;
                        StartAngleSlider.Minimum = selectedPart.MinimumAngle;
                        StartAngleSlider.Maximum = selectedPart.MaximumAngle;
                        StartAngleSlider.Value = selectedPart.CurrentAngle;
                        FirstJointLabel.Content = selectedPart.Name;

                        if (selectedPart.RelatedParts.Count > 0)
                        {
                            StartAngle2Slider.Minimum = selectedPart.RelatedParts[0].MinimumAngle;
                            StartAngle2Slider.Maximum = selectedPart.RelatedParts[0].MaximumAngle;
                            StartAngle2Slider.Value = selectedPart.RelatedParts[0].CurrentAngle;
                            StartAngle2Slider.Visibility = Visibility.Visible;
                            StartAngle2Text.Visibility = Visibility.Visible;

                            SecondJointLabel.Content = selectedPart.RelatedParts[0].Name;
                            SecondJointLabel.Visibility = Visibility.Visible;
                        }
                        else
                        {
                            StartAngle2Slider.Visibility = Visibility.Collapsed;
                            StartAngle2Text.Visibility = Visibility.Collapsed;
                            SecondJointLabel.Visibility = Visibility.Collapsed;
                        }

                        if (selectedPart.RelatedParts.Count > 1)
                        {
                            StartAngle3Slider.Minimum = selectedPart.RelatedParts[1].MinimumAngle;
                            StartAngle3Slider.Maximum = selectedPart.RelatedParts[1].MaximumAngle;
                            StartAngle3Slider.Value = selectedPart.RelatedParts[1].CurrentAngle;
                            StartAngle3Slider.Visibility = Visibility.Visible;
                            StartAngle3Text.Visibility = Visibility.Visible;

                            ThirdJointLabel.Content = selectedPart.RelatedParts[1].Name;
                            ThirdJointLabel.Visibility = Visibility.Visible;
                        }
                        else
                        {
                            StartAngle3Slider.Visibility = Visibility.Collapsed;
                            StartAngle3Text.Visibility = Visibility.Collapsed;
                            ThirdJointLabel.Visibility = Visibility.Collapsed;
                        }
                        slidersInitilization = false;

                        StartAngleText.Text = StartAngleSlider.Value.ToString();
                        StartAngle2Text.Text = StartAngle2Slider.Value.ToString();
                        StartAngle3Text.Text = StartAngle3Slider.Value.ToString();

                        return true;
                    }
                    selectedModel = hitModel;
                }
            }
            return false;
        }

        private void RotateRobot()
        {            
            DoubleAnimation angleAnimation = new DoubleAnimation();
            angleAnimation.From = 0;
            angleAnimation.To = 360;
            angleAnimation.Duration = TimeSpan.FromSeconds(2);                        

            AxisAngleRotation3D rotation = new AxisAngleRotation3D(
                new Vector3D(1, 0, 0),
                10);

            selectedModel.Transform = new RotateTransform3D(rotation, selectedModel.Content.Bounds.Location);
            rotation.BeginAnimation(AxisAngleRotation3D.AngleProperty, angleAnimation);
        }

        private void RotateSelectedJoint(double angle, double angle2, double angle3)
        {
            if (selectedModel != null)
            {
                selectedPart.SetRotation(angle);

                if (selectedPart.RelatedParts.Count > 0)
                    selectedPart.RelatedParts[0].SetRotation(angle2);
                if (selectedPart.RelatedParts.Count > 1)
                    selectedPart.RelatedParts[1].SetRotation(angle3);

                //robot3d.UpdatePartPositions();
            }
        }

        private void StartAngleSlider_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            if (!slidersInitilization)
            {
                StartAngleSlider.Value = Math.Round(StartAngleSlider.Value, 0);
                StartAngleText.Text = StartAngleSlider.Value.ToString();
                RotateSelectedJoint(StartAngleSlider.Value, StartAngle2Slider.Value, StartAngle3Slider.Value);
            }
        }         
        
        private void StartAngleText_TextChanged(object sender, TextChangedEventArgs e)
        {
            int angle = 0;
            if (int.TryParse(StartAngleText.Text, out angle))
                StartAngleSlider.Value = angle;
        }

        private void StartAngle2Slider_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            if (!slidersInitilization)
            {
                StartAngle2Slider.Value = Math.Round(StartAngle2Slider.Value, 0);
                StartAngle2Text.Text = StartAngle2Slider.Value.ToString();
                RotateSelectedJoint(StartAngleSlider.Value, StartAngle2Slider.Value, StartAngle3Slider.Value);
            }
        }

        private void StartAngle2Text_TextChanged(object sender, TextChangedEventArgs e)
        {
            int angle = 0;
            if (int.TryParse(StartAngle2Text.Text, out angle))
                StartAngle2Slider.Value = angle;
        }

        private void StartAngle3Text_TextChanged(object sender, TextChangedEventArgs e)
        {
            int angle = 0;
            if (int.TryParse(StartAngle3Text.Text, out angle))
                StartAngle3Slider.Value = angle;
        }

        private void StartAngle3Slider_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            if (!slidersInitilization)
            {
                StartAngle3Slider.Value = Math.Round(StartAngle3Slider.Value, 0);
                StartAngle3Text.Text = StartAngle3Slider.Value.ToString();
                RotateSelectedJoint(StartAngleSlider.Value, StartAngle2Slider.Value, StartAngle3Slider.Value);
            }
        }

        private void AddStepButton_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                int duration = int.Parse(DurationText.Text);

                // Zisti aktualnu poziciu
                RobotAnimator animator = new RobotAnimator(robot3d);
                animator.SetJointMotions(jointMotions);
                Robot3DState selState = animator.Animation.GetStateInTime(Robot3DState.StartState, AnimationSlider.Value);

                //Robot3DState lastState = Robot3DState.GetState(Robot3DState.StartState, jointMotions);
                List<JointMotion> motions = Robot3DState.GetJointMotions(selState, new Robot3DState(robot3d), (int)AnimationSlider.Value, duration);

                for (int i = 0; i < motions.Count; i++)
                {
                    StepList.Items.Add(motions[i]);
                }

                jointMotions.AddRange(motions);

                animator.SetJointMotions(motions);
                AnimationSlider.Maximum = animator.Animation.GetLastMotion().StartTime + animator.Animation.GetLastMotion().Duration;
                AnimationSlider.Value = AnimationSlider.Maximum;
            }
            catch (OverflowException)
            {
                MessageBox.Show("Hodnota trvania kroku je príliš veľká.", null, MessageBoxButton.OK, MessageBoxImage.Information);
            }
            catch (FormatException)
            {
                MessageBox.Show("Hodnota trvania kroku má nesprávny formát.", null, MessageBoxButton.OK, MessageBoxImage.Information);
            }
            catch (Exception ex)
            {
                MessageBox.Show("Krok nebolo možné pridať kvôli neočakávanej chybe.\n\n" + ex.ToString(),
                    null, MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        private void DeleteStepButton_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                StepList.Items.RemoveAt(StepList.SelectedIndex);
            }
            catch
            {
            }
        }

        private void PlayButton_Click(object sender, RoutedEventArgs e)
        {
            if (animator.IsBusy)
                animator.Stop();
            else
            {
                animator.SetJointMotions(jointMotions);
                animator.Start(robotViewport.Dispatcher, AnimationSlider);
            }
        }
        
        private void UpdateStepList()
        {
            StepList.Items.Clear();
            for (int i = 0; i < jointMotions.Count; i++)
            {
                StepList.Items.Add(jointMotions[i]);
            }
        }

        public System.Drawing.Bitmap GetRobotMotionPreview(int time)
        {            
            System.Drawing.Bitmap retval = null;

            this.Dispatcher.Invoke(DispatcherPriority.Normal, (Action)(() =>
            {
                SetRobotPosition(time);

                RenderTargetBitmap bmp = new RenderTargetBitmap(200, 200, 96, 96, PixelFormats.Pbgra32);
                bmp.Render(robotViewport);

                PngBitmapEncoder encoder = new PngBitmapEncoder();
                encoder.Frames.Add(BitmapFrame.Create(bmp));

                MemoryStream ms = new MemoryStream();
                encoder.Save(ms);

                retval = new System.Drawing.Bitmap(ms);
            }));

            return retval;
        }

        private void SetRobotPosition(int time)
        {
            Robot3DState state = (new RobotAnimation(jointMotions).GetStateInTime(Robot3DState.StartState, time));
            robot3d.SetRobotState(state);            
                        
        }

        private void AnimationSlider_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)        
        {
            if (!animator.IsBusy)
                SetRobotPosition((int)(AnimationSlider.Value));

            //PhysicalModel physics = new PhysicalModel(robot3d, new Gyroscope(0, 0, 0));
        }
    }
}
